/*
 * metadata/gc.c: GC icalls.
 *
 * Author: Paolo Molaro <lupus@ximian.com>
 *
 * (C) 2002 Ximian, Inc.
 */

#include <config.h>
#include <glib.h>
#include <string.h>

#include <mono/metadata/gc.h>
#include <mono/metadata/threads.h>
#if HAVE_BOEHM_GC
#define GC_I_HIDE_POINTERS
#include <gc/gc.h>
#else
#define HIDE_POINTER(v) (v)
#define REVEAL_POINTER(v) (v)
#endif

static int finalize_slot = -1;

/* 
 * actually, we might want to queue the finalize requests in a separate thread,
 * but we need to be careful about the execution domain of the thread...
 */
static void
run_finalize (void *obj, void *data)
{
	MonoObject *exc = NULL;
	MonoObject *o;
	o = (MonoObject*)((char*)obj + GPOINTER_TO_UINT (data));

	if (finalize_slot < 0) {
		int i;
		for (i = 0; i < mono_defaults.object_class->vtable_size; ++i) {
			MonoMethod *cm = mono_defaults.object_class->vtable [i];
	       
			if (!strcmp (cm->name, "Finalize")) {
				finalize_slot = i;
				break;
			}
		}
	}
	/* speedup later... */
	/* g_print ("Finalize run on %s\n", mono_object_class (o)->name); */
	mono_runtime_invoke (o->vtable->klass->vtable [finalize_slot], o, NULL, &exc);

	if (exc) {
		/* fixme: do something useful */
	}
}

/*
 * Some of our objects may point to a different address than the address returned by GC_malloc()
 * (because of the GetHashCode hack), but we need to pass the real address to register_finalizer.
 * This also means that in the callback we need to adjust the pointer to get back the real
 * MonoObject*.
 * We also need to be consistent in the use of the GC_debug* variants of malloc and register_finalizer, 
 * since that, too, can cause the underlying pointer to be offset.
 */
static void
object_register_finalizer (MonoObject *obj, void (*callback)(void *, void*))
{
#if HAVE_BOEHM_GC
	guint offset = 0;

	g_assert (GC_base (obj) == (char*)obj - offset);
	GC_register_finalizer ((char*)obj - offset, callback, GUINT_TO_POINTER (offset), NULL, NULL);
#endif
}

void
mono_object_register_finalizer (MonoObject *obj)
{
	object_register_finalizer (obj, run_finalize);
}

void
ves_icall_System_GC_InternalCollect (int generation)
{
#if HAVE_BOEHM_GC
	GC_gcollect ();
#endif
}

gint64
ves_icall_System_GC_GetTotalMemory (MonoBoolean forceCollection)
{
#if HAVE_BOEHM_GC
	if (forceCollection)
		GC_gcollect ();
	return GC_get_heap_size ();
#else
	return 0;
#endif
}

void
ves_icall_System_GC_KeepAlive (MonoObject *obj)
{
	/*
	 * Does nothing.
	 */
}

void
ves_icall_System_GC_ReRegisterForFinalize (MonoObject *obj)
{
	object_register_finalizer (obj, run_finalize);
}

void
ves_icall_System_GC_SuppressFinalize (MonoObject *obj)
{
	object_register_finalizer (obj, NULL);
}

void
ves_icall_System_GC_WaitForPendingFinalizers (void)
{
}

/*static CRITICAL_SECTION handle_section;*/
static guint32 next_handle = 0;
static gpointer *gc_handles = NULL;
static guint32 array_size = 0;

/*
 * The handle type is encoded in the lower two bits of the handle value:
 * 0 -> normal
 * 1 -> pinned
 * 2 -> weak
 */

/*
 * FIXME: make thread safe and reuse the array entries.
 */
MonoObject *
ves_icall_System_GCHandle_GetTarget (guint32 handle)
{
	MonoObject *obj;

	if (gc_handles) {
		obj = gc_handles [handle >> 2];
		if ((handle & 0x3) > 1)
			return REVEAL_POINTER (obj);
		return obj;
	}
	return NULL;
}

guint32
ves_icall_System_GCHandle_GetTargetHandle (MonoObject *obj, guint32 handle, gint32 type)
{
	gpointer val = obj;
	guint32 h, idx = next_handle++;

	if (idx >= array_size) {
#if HAVE_BOEHM_GC
		gpointer *new_array;
		if (!array_size)
			array_size = 16;
		new_array = GC_malloc (sizeof (gpointer) * (array_size * 2));
		if (gc_handles) {
			int i;
			memcpy (new_array, gc_handles, sizeof (gpointer) * array_size);
			/* need to re-register links for weak refs. test if GC_realloc needs the same */
			for (i = 0; i < array_size; ++i) {
				if (((gulong)new_array [i]) & 0x1) { /* all and only disguised pointers have it set */
					GC_general_register_disappearing_link (&(new_array [i]), REVEAL_POINTER (new_array [i]));
				}
			}
		}
		array_size *= 2;
		gc_handles = new_array;
#else
		g_error ("No GCHandle support built-in");
#endif
	}
	h = idx << 2;

	/* resuse the type from the old target */
	if (type == -1)
		type =  handle & 0x3;
	switch (type) {
	case 0:
	case 1:
		h |= type;
		gc_handles [idx] = val;
		break;
	default:
		h |= 2;
		val = (gpointer)HIDE_POINTER (val);
		gc_handles [idx] = val;
#if HAVE_BOEHM_GC
		GC_general_register_disappearing_link (&(gc_handles [idx]), obj);
#else
		g_error ("No weakref support");
#endif
		break;
	}
	return h;
}

void
ves_icall_System_GCHandle_FreeHandle (guint32 handle)
{
	gc_handles [handle >> 2] = (gpointer)-1;
}

gpointer
ves_icall_System_GCHandle_GetAddrOfPinnedObject (guint32 handle)
{
	MonoObject *obj;

	if (gc_handles) {
		obj = gc_handles [handle >> 2];
		if ((handle & 0x3) > 1)
			return REVEAL_POINTER (obj);
		return obj;
	}
	return NULL;
}


